<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2019 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2019.1/1829459
 */

namespace Projects\Model;

use Application\Filter\ArrayValues;
use Application\View\Helper\ViewHelperFactory;
use Groups\Model\Group;
use P4\Connection\ConnectionInterface as Connection;
use P4\Connection\Exception\ServiceNotFoundException;
use P4\File\Filter;
use P4\File\Query;
use P4\File\File;
use P4\Key\Key;
use P4\Model\Fielded\Iterator;
use P4\Model\Connected\Iterator as ConnectedIterator;
use P4\Spec\Change;
use P4\Spec\Client;
use P4\Spec\Job;
use P4\Spec\Exception\NotFoundException as P4SpectNotFoundException;
use Record\Exception\NotFoundException as RecordNotFoundException;
use Record\Exception\NotFoundException;
use Record\Key\AbstractKey;
use Users\Model\User;
use Users\Model\Config as UserConfig;

class Project extends AbstractKey
{
    const KEY_PREFIX = 'swarm-project-';

    const FETCH_BY_MEMBER            = 'member';
    const FETCH_BY_WORKFLOW          = 'workflow';
    const FETCH_BY_BRANCHES_WORKFLOW = 'branches';
    const FETCH_COUNT_FOLLOWERS      = 'countFollowers';
    const FETCH_INCLUDE_DELETED      = 'includeDeleted';
    const FETCH_NO_CACHE             = 'noCache';

    const MEMBERSHIP_LEVEL_OWNER     = 'owner';
    const MEMBERSHIP_LEVEL_MEMBER    = 'member';
    const MEMBERSHIP_LEVEL_FOLLOWER  = 'following';
    const MEMBERSHIP_LEVEL_MODERATOR = 'moderator';

    const FORMAT_URL  = 'URL';
    const FORMAT_JSON = 'JSON';

    const FIELD_NAME                     = 'name';
    const FIELD_DESCRIPTION              = 'description';
    const FIELD_WORKFLOW                 = 'workflow';
    const FIELD_RETAIN_DEFAULT_REVIEWERS = 'retainDefaultReviewers';
    const FIELD_MINIMUM_UP_VOTES         = 'minimumUpVotes';

    const MAX_VALUE = 1000;

    protected $needsGroup = false;
    protected $fields     = array(
        Project::FIELD_NAME                     => array(
            'accessor'  => 'getName',
            'mutator'   => 'setName'
        ),
        'defaults'                              => array(
            'accessor'  => 'getDefaults',
            'mutator'   => 'setDefaults'
        ),
        Project::FIELD_DESCRIPTION              => array(
            'accessor'  => 'getDescription',
            'mutator'   => 'setDescription'
        ),
        'members'       => array(
            'accessor'  => 'getMembers',
            'mutator'   => 'setMembers'
        ),
        'subgroups'     => array(
            'accessor'  => 'getSubgroups',
            'mutator'   => 'setSubgroups',
            'unstored'  => true
        ),
        'owners'        => array(
            'accessor'  => 'getOwners',
            'mutator'   => 'setOwners',
        ),
        'branches'      => array(
            'index'     => 1602,
            'indexer'   => 'getBranchWorkflows',
            'accessor'  => 'getBranches',
            'mutator'   => 'setBranches'
        ),
        'jobview'       => array(
            'accessor'  => 'getJobview',
            'mutator'   => 'setJobview'
        ),
        'emailFlags'    => array(
            'accessor'  => 'getEmailFlags',
            'mutator'   => 'setEmailFlags'
        ),
        'tests'         => array(
            'accessor'  => 'getTests',
            'mutator'   => 'setTests',
            'hidden'    => true
        ),
        'deploy'        => array(
            'accessor'  => 'getDeploy',
            'mutator'   => 'setDeploy',
            'hidden'    => true
        ),
        'deleted'                               => array(
            'accessor'  => 'isDeleted',
            'mutator'   => 'setDeleted'
        ),
        'private'                               => array(
            'accessor'  => 'isPrivate',
            'mutator'   => 'setPrivate'
        ),
        Project::FIELD_WORKFLOW                 => array(
            'index'         => 1601,
        ),
        Project::FIELD_RETAIN_DEFAULT_REVIEWERS => array(
            'accessor'  => 'areDefaultReviewersRetained',
            'mutator'   => 'setRetainDefaultReviewers'
        ),
        Project::FIELD_MINIMUM_UP_VOTES         => array(
            'accessor'  => 'getMinimumUpVotes',
            'mutator'   => 'setMinimumUpVotes'
        )
    );

    private static $definedFieldsCache = null;
    // Define default roles that will be looked for to establish if a user is interested in a project
    private static $projectRoles = array(
        Project::MEMBERSHIP_LEVEL_MEMBER,
        Project::MEMBERSHIP_LEVEL_OWNER,
        Project::MEMBERSHIP_LEVEL_FOLLOWER,
        Project::MEMBERSHIP_LEVEL_MODERATOR
    );
    // A cache of projects that have already been retrieved
    const  L2_CACHE_ENABLED = true;  // Here for dev testing
    private static $l2Cache = array();

    /**
     * Verifies if the specified record(s) exists.
     * Its better to call 'fetch' directly in a try block if you will
     * be retrieving the record on success.
     *
     * @param   string|int|array    $id     the entry id or an array of ids to filter
     * @param   Connection          $p4     the connection to use
     * @return  bool|array          true/false for single arg, array of existent ids for array input
     * @throws \Exception
     */
    public static function exists($id, Connection $p4)
    {
        $found = array();
        foreach ((array)$id as $projectId) {
            try {
                $found[] = Project::fetch($projectId, $p4)->getId();
            } catch (RecordNotFoundException $nfe) {
                // Non-existent, don't mark as found
            } catch (\InvalidArgumentException $iae) {
                // Bad id, don't mark as found
            }
        }
        return is_array($id) ? $found : count($found) != 0;
    }

    /**
     * Get all of the field names that are defined in this model (excludes ad-hoc fields).
     * Overridden to only get fields if not already statically set
     *
     * @return  array   a list of defined field names for this model.
     */
    public function getDefinedFields()
    {
        // field names are taken from the fields array
        // if the element value is an array, we assume the key is the
        // field name; otherwise, we assume the value is the field name.
        if (!static::$definedFieldsCache) {
            foreach ($this->fields as $key => $value) {
                static::$definedFieldsCache[] = is_array($value) ? $key : $value;
            }
        }
        return static::$definedFieldsCache;
    }

    /**
     * Extends fetch to use cache if available.
     * Note: Deleted projects are not included in the result.
     *
     * @param   string          $id         the id of the entry to fetch
     * @param   Connection      $p4         a specific connection to use
     * @return  Project                     instance of the requested entry
     * @throws  RecordNotFoundException     if project with the given id doesn't exist
     */
    public static function fetch($id, Connection $p4)
    {
        if (!isset(Project::$l2Cache[$id]) || !Project::L2_CACHE_ENABLED) {
            Project::$l2Cache[$id] = parent::fetch($id, $p4);
        }
        $project = Project::$l2Cache[$id];

        // if we have a project and it is not deleted, return it
        if (isset($project) && !$project->isDeleted()) {
            $clone = clone $project;
            return $clone->setConnection($p4);
        }

        throw new RecordNotFoundException("Cannot fetch entry. Id does not exist.");
    }

    /**
     * Extends parent to add additional options (listed below) and to use cache if available.
     * To simplify the code, we support only a subset of options that are available in parent.
     * By default, deleted projects will not be included in the result. To include them,
     * FETCH_INCLUDE_DELETED option with value set to true must be passed in options.
     *
     * @param   array       $options    currently supported options are:
     *                                        FETCH_BY_IDS - provide an array of ids to fetch
     *                                     FETCH_BY_MEMBER - set to limit results to include only projects
     *                                                       having the given member
     *                                   FETCH_BY_WORKFLOW - set to limit results to include only projects
     *                                                       and their branches for the given workflow. This
     *                                                       can NOT be used in conjunction with any other option.
     *                               FETCH_COUNT_FOLLOWERS - if true, each project will include a 'followers'
     *                                                       flag indicating the number of followers
     *                               FETCH_INCLUDE_DELETED - set to true to also include deleted projects
     *                                      FETCH_NO_CACHE - set to true to avoid using the cache
     * @param   Connection  $p4             the perforce connection to use
     * @return  Iterator|array              the list of zero or more matching project objects | an array of
     *                                      projects that are linked to a given workflow.
     * @throws  \InvalidArgumentException   if the caller passed option(s) we don't support
     * @throws  \Exception                  for unexpected issues
     */
    public static function fetchAll(array $options, Connection $p4)
    {
        // prepare default values for supported options
        $defaults = array(
            static::FETCH_BY_IDS          => null,
            static::FETCH_BY_MEMBER       => null,
            static::FETCH_BY_WORKFLOW     => null,
            static::FETCH_COUNT_FOLLOWERS => null,
            static::FETCH_INCLUDE_DELETED => false,
            static::FETCH_NO_CACHE        => null,
            static::FETCH_BY_KEYWORDS     => null,
            static::FETCH_KEYWORDS_FIELDS => null,
            static::FETCH_MAXIMUM         => null,
            static::FETCH_AFTER           => null
        );

        // throw if user passed option(s) we don't support
        $unsupported = array_diff(array_keys($options), array_keys($defaults));
        if (count($unsupported)) {
            throw new \InvalidArgumentException(
                'Following option(s) are not valid for fetching projects: ' . implode(', ', $unsupported) . '.'
            );
        }

        $options += $defaults;

        // Return early if there is a fetch by id request for no data
        if (isset($options[static::FETCH_BY_IDS]) && 0 === count($options[static::FETCH_BY_IDS])) {
            return new Iterator();
        }

        // fetch all projects, try to get them from cache if possible
        if (!$options[static::FETCH_BY_WORKFLOW] && !$options[static::FETCH_NO_CACHE] &&
            !($options[static::FETCH_MAXIMUM] && !$options[static::FETCH_AFTER]) &&
            !($options[static::FETCH_BY_IDS] || 75<count((array)$options[static::FETCH_BY_IDS]))) {
            try {
                $cached = static::getCachedProjects($p4);
                // We need to take care of FETCH_BY_IDS option that is otherwise handled by parent
                // Do this before all the clone and connection setting work to limit the number
                // Use FILTER_COPY to avoid modifying the Iterator kept in the cache directly
                if ($options[static::FETCH_BY_IDS] !== null) {
                    $cached = $cached->filter(
                        'id',
                        (array) $options[static::FETCH_BY_IDS],
                        array(ConnectedIterator::FILTER_COPY)
                    );
                }
                $startAt = $options[static::FETCH_AFTER];
                $stopAt  = $options[static::FETCH_MAXIMUM];
                $started = false;
                // projects from cache need to be cloned and given a connection
                $projects = new Iterator;
                foreach ($cached as $key => $project) {
                    if ($startAt && !$started) {
                        $started = $project->getId() === $startAt;
                        continue;
                    }
                    if ($stopAt && count($projects) === $stopAt) {
                        // Stop iterating if we have enough data
                        break;
                    }
                    $projects[$key] = clone $project;
                    $projects[$key]->setConnection($p4);
                }
            } catch (ServiceNotFoundException $e) {
                // not using cache - projects will be populated later
            }
        }

        // For a single FETCH_BY_IDS or FETCH_BY_NAME, revert to fetch unless DELETED projects are needed
        if ((isset($options[Key::FETCH_BY_NAME]) || count((array)$options[static::FETCH_BY_IDS]) === 1) &&
            !$options[static::FETCH_INCLUDE_DELETED]) {
            $projects = new Iterator();
            $id       = isset($options[Key::FETCH_BY_NAME])
                ? $options[Key::FETCH_BY_NAME]
            : current((array)$options[static::FETCH_BY_IDS]);
            try {
                $project                     = static::fetch($id, $p4);
                $projects[$project->getId()] = $project;
            } catch (NotFoundException $nfe) {
                // FetchAll allows data to not exist, with an empty array returned.
            }
        }

        // Projects do not have any indexed fields - but we can still support filter by keywords as long as we unset
        // the values so that the super class does not build an empty search expression. We support it here in the
        // model to use the in-built 'filter' method rather than the controller doing its own thing
        $keywords       = $options[static::FETCH_BY_KEYWORDS];
        $keywordsFields = $options[static::FETCH_KEYWORDS_FIELDS];
        unset($options[static::FETCH_BY_KEYWORDS]);
        unset($options[static::FETCH_KEYWORDS_FIELDS]);

        $workflow = $options[static::FETCH_BY_WORKFLOW];
        if ($workflow) {
            // At present we need to search by only branches. So including them automatically as part of the search.
            $options[static::FETCH_BY_BRANCHES_WORKFLOW] = isset($options[static::FETCH_BY_BRANCHES_WORKFLOW])
                ?:$workflow;
            $options[static::FETCH_SEARCH]               = static::makeSearchExpression($options);
        }

        // get projects from parent if either user requested fetching with no cache or cache is not available
        $projects = isset($projects) ? $projects : parent::fetchAll($options, $p4);

        if ($keywords && $keywordsFields) {
            $projects->filter($keywordsFields, $keywords, array(Iterator::FILTER_CONTAINS, Iterator::FILTER_NO_CASE));
        }

        // unless explicitly requested, filter out deleted projects
        if (!$options[static::FETCH_INCLUDE_DELETED]) {
            $projects->filter('deleted', false);
        }

        // handle FETCH_BY_MEMBER
        $member = $options[static::FETCH_BY_MEMBER];
        if ($member) {
            $projects->filterByCallback(
                function (Project $project) use ($member) {
                    return $project->isMember($member);
                }
            );
        }

        // if caller requested follower counts, add them now.
        if ($options[static::FETCH_COUNT_FOLLOWERS]) {
            $followers = UserConfig::fetchFollowerCounts(
                array(UserConfig::COUNT_BY_TYPE => 'project'),
                $p4
            );

            foreach ($projects as $project) {
                $key = 'project:' . $project->getId();
                $project->set('followers', isset($followers[$key]) ? $followers[$key]['count'] : 0);
            }
        }

        return $projects;
    }

    /**
     * Produces a 'p4 search' expression for the given field/value pairs.
     *
     * Extends parent to put the branches and project level workflow search
     * together.
     *
     * @param   array   $conditions     field/value pairs to search for
     *
     * @return  string  a query expression suitable for use with p4 search
     */
    protected static function makeSearchExpression($conditions)
    {
        $workflowSearches = array();
        // First find the workflow searches and remove them from the parent expression builder
        foreach (array(Project::FETCH_BY_WORKFLOW, Project::FETCH_BY_BRANCHES_WORKFLOW) as $condition) {
            if (isset($conditions[$condition])) {
                $workflowSearches[] = parent::makeSearchExpression(array($condition => $conditions[$condition]));
                unset($conditions[$condition]);
            }
        }
        $workflowSearchExpression = implode('|', $workflowSearches);

        // Now build the expression for any other conditions that we have passed in.
        $expression = parent::makeSearchExpression($conditions);
        // Put both the workflow search and any other searches together.
        $expression = $expression . $workflowSearchExpression;

        return $expression;
    }

    /**
     * Attempts to clear cache for projects.
     *
     * @param   Connection  $p4     specific connection to use
     */
    public static function clearCache(Connection $p4)
    {
        try {
            Project::$l2Cache = array();
            $cache            = $p4->getService('cache');
            $cache->invalidateItem('projects');
        } catch (ServiceNotFoundException $e) {
            // no cache - nothing to invalidate
        }
    }

    /**
     * The friendly name for this project.
     *
     * @return  string|null     the name for this project.
     */
    public function getName()
    {
        return $this->getRawValue(Project::FIELD_NAME);
    }

    /**
     * Set a friendly name for this project.
     *
     * @param   string|null     $name   the friendly name for this project or null
     * @return  Project         to maintain a fluent interface
     */
    public function setName($name)
    {
        return $this->setRawValue(Project::FIELD_NAME, $name);
    }

    /**
     * The defaults for this project, always provided, can be empty.
     *
     * @return  array     the default values for this project.
     */
    public function getDefaults()
    {
        $defaults =  (array)$this->getRawValue('defaults') + array('reviewers' => array());
        ksort($defaults['reviewers']);
        return $defaults;
    }

    /**
     * Set the defaults for this project.
     *
     * @param   array     $defaults    an array of name/value pairs
     * @return  Project   to maintain a fluent interface
     */
    public function setDefaults(array $defaults)
    {
        return $this->setRawValue('defaults', $this->normaliseDefaults($defaults));
    }

    /**
     * Process a project/branch defaults value, applying normalisation as neccesary.
     * Reviewers will have normalised values for the required attribute; true, "1" or unset
     * @param array $defaults
     * @return array normalised $defaults
     */
    private function normaliseDefaults(array $defaults)
    {
        foreach ($defaults as $default => &$values) {
            if ($default === 'reviewers') {
                foreach ($values as &$reviewer) {
                    if (isset($reviewer['required']) && $reviewer['required'] !== "1") {
                        if ("false" === $reviewer['required']) {
                            unset($reviewer['required']);
                        } else {
                            $reviewer['required'] = true;
                        }
                    }
                }
            }
        }
        return $defaults;
    }
    /**
     * The description for this project.
     *
     * @return  string|null     the description for this project.
     */
    public function getDescription()
    {
        return $this->getRawValue(Project::FIELD_DESCRIPTION);
    }

    /**
     * Set a description for this project.
     *
     * @param   string|null     $description    the description for this project or null
     * @return  Project         to maintain a fluent interface
     */
    public function setDescription($description)
    {
        return $this->setRawValue(Project::FIELD_DESCRIPTION, $description);
    }

    /**
     * Returns an array of member ids associated with this project.
     * @return  array   ids of all members for this project
     */
    public function getMembers()
    {
        $this->loadGroup();
        return $this->getSortedField('members');
    }

    /**
     * Returns an array of subgroups that are members of this project
     * (i.e. subgroups of the project group).
     *
     * @return  array   ids of all subgroups under this project
     */
    public function getSubgroups()
    {
        $this->loadGroup();
        return (array) $this->getRawValue('subgroups');
    }

    /**
     * Get all members of this project recursively.
     * @param bool $flip if true array keys are the user ids (default is false)
     * @return array|null flat list of all members
     * @throws \P4\Exception
     */
    public function getAllMembers($flip = false)
    {
        return Group::fetchAllMembers($this->id, $flip, null, null, $this->getConnection());
    }

    /**
     * Gets the raw field value. Will sort alphabetically unless flip is requested.
     * @param $field
     * @param bool $flip
     * @return array|null
     */
    private function getSortedField($field, $flip = false)
    {
        $values = $flip
            ? array_flip((array) $this->getRawValue($field))
            : (array) $this->getRawValue($field);
        if (!$flip) {
            sort($values, SORT_STRING);
        }
        return $values;
    }

    /**
     * Get all Users and groups of this project
     * @param bool $flip if true array keys are the user ids (default is false)
     * @return array|null flat list of all members
     * @throws \P4\Exception
     */
    public function getUsersAndSubgroups($flip = false)
    {
        return Group::fetchUsersAndSubgroups($this->id, $this->getConnection());
    }

    /**
     * Returns an array of owner ids associated with this project.
     * @param   bool    $flip       if true array keys are the user ids (default is false)
     * @return  array   ids of all owners for this project
     */
    public function getOwners($flip = false)
    {
        return $this->getSortedField('owners', $flip);
    }

    /**
     * Returns true if this project has one or more owners and false otherwise.
     *
     * @return  bool    true if project has at least one owner, false otherwise
     */
    public function hasOwners()
    {
        return count($this->getOwners()) > 0;
    }

    /**
     * Set owners for this project.
     *
     * @param   array|null  the owners for this project
     * @return  Project     to maintain a fluent interface
     */
    public function setOwners($owners)
    {
        return $this->setRawValue('owners', $owners);
    }

    /**
     * Get a list of users that follow this project.
     * Members are implicitly followers but are not listed by default.
     *
     * @param   bool|array  $excludeMembers     optional - exclude members (defaults to true)
     *                                          the list of members can be given (useful for performance)
     * @return  array       a list of ids of users that are following this project
     * @throws \P4\Exception
     */
    public function getFollowers($excludeMembers = true)
    {
        $followers = UserConfig::fetchFollowerIds(
            $this->getId(),
            'project',
            $this->getConnection()
        );

        // optionally exclude members
        if ($excludeMembers) {
            $members   = is_array($excludeMembers) ? $excludeMembers : $this->getAllMembers();
            $followers = array_diff($followers, $members);
        }

        sort($followers, SORT_STRING);
        return $followers;
    }

    /**
     * Set an array of member ids for this project.
     *
     * @param   array|null  $members    an array of members or null
     * @return  Project     to maintain a fluent interface
     */
    public function setMembers($members)
    {
        return $this->setRawValue('members', $members);
    }

    /**
     * Set an array of subgroup ids for this project.
     *
     * @param   array|null  $groups     an array of groups or null
     * @return  Project     to maintain a fluent interface
     * @throws  RuntimeException            if the server is too old to manage groups
     * @throws  InvalidArgumentException    if groups contains project id
     */
    public function setSubgroups($groups)
    {
        if (in_array($this->id, $groups, true)) {
            throw new \InvalidArgumentException("Cannot set project as a subgroup of itself.");
        }

        return $this->setRawValue('subgroups', $groups);
    }

    /**
     * The resulting array with entries for all known branches.
     *
     * This will be an array of arrays. Each sub-array should contain
     * keys for: id, name, paths and moderators.
     *
     * @param   string|null     $sortField  optional - field to sort branches list on (using natural,
     *                                      case-insensitive sort)
     * @param   array           $mainlines  optional - branch names to appear on top of the list when sorted
     * @return  array           the branches for this project
     */
    public function getBranches($sortField = null, array $mainlines = array())
    {
        if (!is_null($sortField) && !is_string($sortField)) {
            throw new \InvalidArgumentException('Invalid $sortField format: $sortField must be a string or null.');
        }

        // normalize the branches array we are about to return.
        // we do this on read as there is the unlikely possibility
        // the data was modified externally.
        $branches = (array) $this->getRawValue('branches');
        foreach ($branches as $id => $branch) {
            $branch             += array('id'                    => null);
            $branch             += array(
                'id'                    => null,
                'name'                  => null,
                'paths'                 => array(),
                'moderators'            => array(),
                'moderators-groups'     => array(),
                'defaults'              => array(),
                Project::FIELD_WORKFLOW => null,
                Project::FIELD_RETAIN_DEFAULT_REVIEWERS => false,
                Project::FIELD_MINIMUM_UP_VOTES => null
            );
            $branch['defaults'] += array('reviewers' => array());
            ksort($branch['defaults']['reviewers']);
            $branch['paths'] = (array) $branch['paths'];
            $branches[$id]   = $branch;
        }

        // sort branches with special handling for mainline branches (will appear first)
        if ($sortField) {
            usort(
                $branches,
                function ($a, $b) use ($sortField, $mainlines) {
                    if (!array_key_exists($sortField, $a) || !array_key_exists($sortField, $b)) {
                        throw new \InvalidArgumentException("Cannot sort branches: branch has no '$sortField' field.");
                    }

                    if (in_array(strtolower($a[$sortField]), $mainlines)) {
                        return -1;
                    } elseif (in_array(strtolower($b[$sortField]), $mainlines)) {
                        return 1;
                    }
                    return strnatcasecmp($a[$sortField], $b[$sortField]);
                }
            );
        }

        return $branches;
    }

    /**
     * Get the branch workflow
     *
     * @return array
     */
    public function getBranchWorkflows()
    {
        return array_map(
            function ($branch) {
                return isset($branch[Project::FIELD_WORKFLOW]) ? $branch[Project::FIELD_WORKFLOW] : '';
            },
            $this->getBranches()
        );
    }

    /**
     * Get a particular branch definition.
     *
     * @param   string  $id     the id of the branch definition to get
     * @return  array   the branch definition (id, name and paths)
     * @throws  \InvalidArgumentException    if no such branch defined
     */
    public function getBranch($id)
    {
        $branches = $this->getBranches();
        foreach ($branches as $branch) {
            if ($branch['id'] === $id) {
                return $branch;
            }
        }

        throw new \InvalidArgumentException("Cannot get branch '$id'. Branch is not defined.");
    }

    /**
     * Get a particular branch definition without an error if the branch is not found.
     *
     * @param   string  $id     the id of the branch definition to get
     * @return array|null       the branch definition or null if the branch is not found
     */
    public function branchExists($id)
    {
        try {
            return $this->getBranch($id);
        } catch (\InvalidArgumentException $noBranchException) {
            // The branch name may have been modified since the caller became aware of
            // it and the caller does not have the new name. For an existence check
            // return null rather than throwing an exception
            return null;
        }
    }

    /**
     * Set a branches array for this project.
     *
     * This should be an array of arrays. Each sub-array should contain
     * keys for: id, default, name, paths and moderators.
     *
     * @param   array|null  $branches   the branches for this project
     * @return  Project     to maintain a fluent interface
     */
    public function setBranches($branches)
    {
        if (!empty($branches)) {
            // First loop though each branch entire we have.
            foreach ($branches as $key => &$branch) {
                if (isset($branch[Project::FIELD_RETAIN_DEFAULT_REVIEWERS])) {
                    $branch[Project::FIELD_RETAIN_DEFAULT_REVIEWERS] =
                        (bool)$branch[Project::FIELD_RETAIN_DEFAULT_REVIEWERS];
                }
                if (isset($branch['moderators'])) {
                    // Get the moderators and ensure unique members. Cast is required for backwards
                    // compatibility with API versions that allowed a string value
                    $originalModerators           = (array) $branch['moderators'];
                    $branches[$key]['moderators'] = array_values(array_unique($originalModerators));
                }

                if (isset($branch['moderators-groups'])) {
                    // Get the moderator groups and ensure unique groups.
                    $originalModeratorsGroups            = $branch['moderators-groups'];
                    $branches[$key]['moderators-groups'] = array_values(array_unique($originalModeratorsGroups));
                }

                if (isset($branch['defaults'])) {
                    $branch['defaults'] = $this->normaliseDefaults($branch['defaults']);
                }
            }
        }
        // Set and return the branches data.
        return $this->setRawValue('branches', $branches);
    }

    /**
     * Get list of moderators from given branches of this project.
     *
     * @param   array|null  $branches   optional - limit branches to collect moderators from
     * @return  array       list of moderators for specified branches
     * @throws \P4\Exception
     */
    public function getModerators(array $branches = null)
    {
        $moderators = array();
        foreach ((array) $this->getRawValue('branches') as $branch) {
            if (is_null($branches) || in_array($branch['id'], $branches)) {
                $branch += array(
                    'moderators'            => array(),
                    'moderators-groups'     => array(),
                );

                $moderators = array_merge($moderators, (array) $branch['moderators']);
                if ($branch['moderators-groups']) {
                    foreach ((array) $branch['moderators-groups'] as $moderatorGroup) {
                        $moderators = array_merge(
                            Group::fetchAllMembers(
                                $moderatorGroup,
                                false,
                                null,
                                null,
                                $this->getConnection()
                            ),
                            $moderators
                        );
                    }
                }
            }
        }
        $uniqueModerators = array_values(array_unique($moderators));
        sort($uniqueModerators, SORT_STRING);
        return $uniqueModerators;
    }

    /**
     * Get list of user and groups as moderators from given branches of this project.
     *
     * @param   array|null  $branches   optional - limit branches to collect moderators from
     * @return  array       list of moderators for specified branches
     */
    public function getModeratorsWithGroups(array $branches = null)
    {
        $moderators['Users']  = array();
        $moderators['Groups'] = array();
        foreach ((array) $this->getRawValue('branches') as $branch) {
            if (is_null($branches) || in_array($branch['id'], $branches)) {
                $branch += array(
                    'moderators'            => array(),
                    'moderators-groups'     => array(),
                );

                $moderators['Users'] = array_merge($moderators['Users'], (array) $branch['moderators']);
                sort($moderators['Users'], SORT_STRING);
                if ($branch['moderators-groups']) {
                    $moderators['Groups'] = array_merge($moderators['Groups'], (array) $branch['moderators-groups']);
                    sort($moderators['Groups'], SORT_STRING);
                }
            }
        }
        return $moderators;
    }


    /**
     * The jobview for this project.
     *
     * @return  string|null     the jobview for this project.
     */
    public function getJobview()
    {
        return $this->getRawValue('jobview');
    }

    /**
     * Set a jobview for this project.
     *
     * @param   string|null     $jobview    the jobview for this project or null
     * @return  Project         to maintain a fluent interface
     */
    public function setJobview($jobview)
    {
        return $this->setRawValue('jobview', $jobview);
    }

    /**
     * Returns an array of email/notification flags set on this project.
     *
     * @return  array   names for all email flags currently set on this project
     */
    public function getEmailFlags()
    {
        return (array) $this->getRawValue('emailFlags');
    }

    /**
     * Returns the value of the specified email flag, if it exists, or null if it could not be found.
     *
     * @param   string      $flag   specific email flag we are looking for
     * @return  mixed|null  value of the flag if found, or null if the flag was not found
     */
    public function getEmailFlag($flag)
    {
        $emailFlags = $this->getEmailFlags();
        return isset($emailFlags[$flag]) ? $emailFlags[$flag] : null;
    }

    /**
     * Set an array of active email/notification flags on this comment.
     *
     * @param   array|null  $flags    an array of flags or null
     * @return  Project     to maintain a fluent interface
     */
    public function setEmailFlags($flags)
    {
        return $this->setRawValue('emailFlags', (array)$flags);
    }

    /**
     * An array containing the keys 'enabled' and 'url'
     * to reflect the test settings.
     *
     * @param   string|null     $key    optional - a specific key to retrieve
     * @return  array|null      an array with keys for enabled and url.
     */
    public function getTests($key = null)
    {
        $values = (array) $this->getRawValue('tests') + array('enabled' => false, 'url' => null);

        // handle 2015.4 to 2016.1 upgrade on the fly
        // - renamed 'postParams' key to 'postBody'
        // - renamed 'postFormat' value 'GET' to 'URL'
        if (array_key_exists('postParams', $values) && !array_key_exists('postBody', $values)) {
            $values['postBody'] = $values['postParams'];
            unset($values['postParams']);
        }
        if (array_key_exists('postFormat', $values) && $values['postFormat'] === 'GET') {
            $values['postFormat'] = static::FORMAT_URL;
        }

        if ($key) {
            return isset($values[$key]) ? $values[$key] : null;
        }

        return $values;
    }

    /**
     * Set tests enabled and url properties.
     *
     * @param   array|null  $tests  array with keys for enabled and url or null
     * @return  Project     to maintain a fluent interface
     */
    public function setTests($tests)
    {
        return $this->setRawValue('tests', $tests);
    }

    /**
     * An array containing the keys 'enabled' and 'url'
     * to reflect the deployment settings.
     *
     * @param   string|null     $key    optional - a specific key to retrieve
     * @return  array|null      an array with keys for enabled and url.
     */
    public function getDeploy($key = null)
    {
        $values = (array) $this->getRawValue('deploy') + array('enabled' => false, 'url' => null);

        if ($key) {
            return isset($values[$key]) ? $values[$key] : null;
        }

        return $values;
    }

    /**
     * Set deployment enabled and url properties.
     *
     * @param   array|null  $deploy     array with keys for enabled and url or null
     * @return  Project     to maintain a fluent interface
     */
    public function setDeploy($deploy)
    {
        return $this->setRawValue('deploy', $deploy);
    }

    /**
     * Boolean value indicating whether this project is deleted.
     *
     * There might be records missing this field (it has been added later).
     * We convert missing values to false indicating the project is considered to
     * be deleted only if the 'deleted' field is present and its value is true.
     *
     * @return  boolean     true if this projects is deleted, false otherwise
     */
    public function isDeleted()
    {
        return (bool) $this->getRawValue('deleted');
    }

    /**
     * Set whether this project is deleted (true) or not (false).
     *
     * @param   boolean     $deleted    pass true to indicate that this project is deleted
     *                                  and false to indicate that this project is active
     * @return  Project     to maintain a fluent interface
     */
    public function setDeleted($deleted)
    {
        return $this->setRawValue('deleted', (bool) $deleted);
    }

    /**
     * Boolean value indicating whether this project is private.
     *
     * There might be records missing this field (it has been added later).
     * We convert missing values to false indicating the project is considered to
     * be public by default.
     *
     * Private projects are visible to members and owners only.
     *
     * @return  boolean     true if this projects is private, false otherwise
     */
    public function isPrivate()
    {
        return (bool) $this->getRawValue('private');
    }

    /**
     * Set whether this project is private (true) or not (false).
     *
     * @param   boolean     $private    pass true to indicate that this project is private
     *                                  and false to indicate that this project is public
     * @return  Project     to maintain a fluent interface
     */
    public function setPrivate($private)
    {
        return $this->setRawValue('private', (bool) $private);
    }

    /**
     * Return client specific for this project.
     * Each branch of this project is mapped as top-level folder in client's view:
     *
     *  <branch-path> //<client-id>/<branch-id>/...
     *
     * At the moment, branches with multiple paths are mapped to the same folder.
     * Client will be created if doesn't exist.
     *
     * @param bool $ignoreCommonPath  Ignore Common path to help get correct change.
     *
     * @return string   name of the client specific to this project
     * @throws \P4\Exception
     */
    public function getClient($ignoreCommonPath = false)
    {
        $p4     = $this->getConnection();
        $client = 'swarm-project-' . $this->getId();
        // prepare view mappings based on the project's branches
        $view = array();
        foreach ($this->getBranches() as $branch) {
            $id = $branch['id'];
            // Get the Common Path
            $commonPath = $ignoreCommonPath ? '//' : File::getCommonPath($branch['paths'], $p4->isCaseSensitive());
            foreach ($branch['paths'] as $path) {
                $this->buildClientView($view, $path, $commonPath, $client, $id);
            }
        }
        return $this->createOrUpdateClient($p4, $client, $view);
    }

    /**
     * Return client specific for this project branch.
     *
     * Client will be created if doesn't exist.
     *
     * @param int  $id                This is the branch ID
     * @param bool $ignoreCommonPath  Ignore Common path to help get correct change.
     *
     * @return mixed
     * @throws \P4\Exception
     */
    public function getBranchClient($id, $ignoreCommonPath = false)
    {
        $p4             = $this->getConnection();
        $client         = 'swarm-project-' . $this->getId();
        $interestBranch = $this->getBranch($id);
        // prepare view mappings based on the project's branches
        $view = array();
        // Get the Common Path
        $commonPath = $ignoreCommonPath ? '//' : File::getCommonPath($interestBranch['paths'], $p4->isCaseSensitive());
        foreach ($interestBranch['paths'] as $path) {
            $this->buildClientView($view, $path, $commonPath, $client, $id);
        }
        return $this->createOrUpdateClient($p4, $client, $view);
    }

    /**
     * Build me the client view for the paths given.
     *
     * @param array  $view        This is the client view that has been built and appended to.
     * @param string $path        The path that we add build a client mapping for.
     * @param string $commonPath  This is the common path for all the paths in this branch.
     * @param string $client      The client name.
     * @param int    $id          The branch ID.
     */
    private function buildClientView(&$view, $path, $commonPath, $client, $id)
    {
        // Find if the path starts with a '+' or '-' sign and remove from path
        preg_match('#^(-|\+){1}//([^/]+)#', $path, $match);
        $sign = null;
        if ($match) {
            $sign = $match[1];
            $path = substr($path, 1);
        }
        // First remove the common Path from the path
        $clientSide = substr($path, strlen($commonPath));
        // Take the current wildcard or file name
        $baseName = basename($path);
        // Check if the clientSide has a forward slash and remove if it does.
        $clientSide = substr($clientSide, 0, 1) !== '/' ? $clientSide : substr($clientSide, 1);
        // Build the client view for this path
        $view[] = '"' . $sign . $path . '" "//' . $client . '/' . $id . '/'
            . ($sign === '+' ? $baseName : $clientSide). '"';
    }

    /**
     * Create or updated existing client with new view.
     *
     * @param Connection $p4     The p4 connection to be used.
     * @param string     $client The client name.
     * @param array      $view   This is the client view that has been built and will be used.
     *
     * @return mixed
     */
    private function createOrUpdateClient($p4, $client, $view)
    {
        // normalize and verify the client view spec
        $data = $p4->run('client', array('-o', $client))->expandSequences()->getData(-1);
        $old  = new Client;
        $old->setView((array) $data['View']);
        $new = new Client;
        $new->setView($view);

        if ($old->getView() != $new->getView() || !isset($data['Update'])) {
            $p4->run(
                'client',
                '-i',
                array(
                    'Host' => '',
                    'View' => $view,
                    'Root' => DATA_PATH . '/tmp'
                ) + $data
            );
        }

        return $client;
    }

    /**
     * Determine which projects are affected by the given job.
     *
     * @param   Job         $job        the job to examine
     * @param   Connection  $p4         the perforce connection to use
     * @return  array       a list of affected projects as values (auto-incrementing keys).
     * @throws \P4\Spec\Exception\Exception
     */
    public static function getAffectedByJob(Job $job, Connection $p4)
    {
        // loop over projects and, for those with a valid job view,
        // see which are impacted by the passed job.
        $projects = static::fetchAll(array(), $p4);
        $affected = array();
        foreach ($projects as $project) {
            // extract the job view and break out the various field=value filter(s) on whitespace
            // we generate a conditions array with field ids as keys and a regex pattern as value
            $matched = false;
            $jobview = trim($project->getJobview());
            $filters = preg_split('/\s+/', $jobview);
            $fields  = array_combine(
                array_map('strtolower', $job->getFields()),
                $job->getFields()
            );
            foreach ($filters as $filter) {
                if (!preg_match('/^([^=()|]+)=([^=()|]+)$/', $filter, $matches)) {
                    continue;
                }

                // we escape the pattern but re-activate originally un-escaped '*'
                // characters as being wildcard matches
                list(, $field, $pattern) = $matches;
                $field                   = strtolower($field);
                $pattern                 = '/^' . preg_quote($pattern, '/') . '$/i';
                $pattern                 = preg_replace('/(^|[^\\\\])\\\\\*/', '$1.*', $pattern);

                // if the job lacks the requested field or pattern doesn't match; skip this project
                // we use the 'fields' array to do a case insensitive lookup of the field name
                if (!isset($fields[$field]) || !preg_match($pattern, $job->get($fields[$field]))) {
                    continue 2;
                }

                $matched = true;
            }

            // only include the project if it matched at least one expression. we don't
            // want projects that lack a job view, or contain only invalid views, to hit.
            if ($matched) {
                $affected[] = $project->getId();
            }
        }

        return $affected;
    }

    /**
     * Get quoted paths for the branch compatible with finding affected projects
     *
     * @param array      $branch      The branch
     * @param array      $cachedPaths The cachedPaths if we want to reduce the amount of times we reuse
     *                                the same path.
     *
     * @return array the quoted paths
     */
    public static function getQuotedPaths($branch, &$cachedPaths)
    {
        // ensure all paths are quoted. based on experiments with p4 client,
        // the \ char is literal and it can't contain double-quotes. As such
        // the simple approach of quote wrapping if it isn't already works.
        return array_map(
            function ($path) use (&$cachedPaths) {
                if (!in_array($path, array_keys((array) $cachedPaths))) {
                    $cachedPaths[ $path ] = Project::quoteString($path);
                }
                return $cachedPaths[ $path ];
            },
            (array) $branch['paths']
        );
    }

    /**
     * Helper method to make sure that a string is bookended with double quotes
     * @param  string      $str         The string to quote
     * @return string the quoted string or the original string if already quoted
     */
    public static function quoteString($str)
    {
        return ($str[0] != '"' || substr($str, -1) != '"') ? $str = '"' . $str . '"' : $str;
    }

    /**
     * Determine which projects (and branches) are affected by the given change.
     * @param Change        $change     the change to examine
     * @param Connection    $p4         the perforce connection to use
     * @param string        $flags      flags to pass to a 'p4 describe' command. Defaults to '-sS'
     * @param array|null    $projects   The list of projects to search
     * @return  array       a list of affected projects as keys with a list of affected
     *                      branches under those keys (as the value)
     * @throws \Exception
     */
    public static function getAffectedByChange(Change $change, Connection $p4, $flags = '-sS', $projects = null)
    {
        // get the list of projects and their branches
        $projects    = $projects !== null ? $projects : static::fetchAll(array(), $p4);
        $affected    = array();
        $inHeader    = true;
        $cachedPaths = array();
        // Using Handlers is slower than running the command and iterate over that data.
        $describe   = $p4->run('describe', array($flags, $change->getId()), null, false)->getData();
        $cachedMaps = array();
        foreach ($describe as $data) {
            $data = trim($data, "\r\n");
            // if we are in the header check to see if we have hit the start
            // of the file list and return that we took care of this block.
            if ($inHeader) {
                if (!strlen($data) || $data === "Affected files ..." || $data === "Shelved files ...") {
                    $inHeader = false;
                }
                continue;
            }
            $file = substr($data, 0, strrpos($data, '#'));
            foreach ($projects as $key => $project) {
                $projectId = $project->getId();
                foreach ((array) $project->getRawValue('branches') as $branch) {
                    // Build up a hash map of P4_Maps to use instead of rebuilding them over and over again.
                    isset($cachedMaps[ $projectId ][ $branch['id'] ]) ? :
                        $cachedMaps[ $projectId ][ $branch['id'] ] = new \P4_Map(
                            Project::getQuotedPaths(
                                $branch,
                                $cachedPaths
                            )
                        );

                    $map = $cachedMaps[ $projectId ][ $branch['id'] ];
                    if ($map->includes($file)) {
                        $affected += array($projectId => array());
                        if (!in_array($branch['id'], $affected[ $projectId ])) {
                            $affected[ $projectId ][] = $branch['id'];
                        }
                    }
                }
            }
        }

        return $affected;
    }

    /**
     * Extends parent to also save the group if server allows admins to do so.
     *
     * @return  Project     to maintain a fluent interface
     * @throws \P4\Exception
     * @throws \P4\Spec\Exception\Exception
     */
    public function save()
    {
        // if admins can manage groups, mark the members field as not being
        // directly stored on the key to avoid data duplication.
        $this->fields['members']['unstored'] = true;

        // if the server is too old for admins to manage groups or the 'members'
        // and 'subgroups' fields have not been populated, we're done
        if ((!array_key_exists('members', $this->values) && !array_key_exists('subgroups', $this->values))
        ) {
            parent::save();

            // clear cache and return
            static::clearCache($this->getConnection());
            return $this;
        }

        // attempt to fetch any existing group with this projects raw id
        $group = false;
        $isAdd = false;
        try {
            $group = Group::fetchById($this->id, $this->getConnection());
        } catch (P4SpectNotFoundException $e) {
            unset($e);
        } catch (\InvalidArgumentException $e) {
            unset($e);
        }

        // early exit if member and subgroups lists haven't changed
        if ($group
            && $group->getUsers() == $this->getRawValue('members')
            && $group->getSubgroups() == $this->getRawValue('subgroups')
        ) {
            parent::save();

            // clear cache and return
            static::clearCache($this->getConnection());
            return $this;
        }

        // if the fetch failed, setup a new group as its an add
        if (!$group) {
            $isAdd = true;
            $group = new Group($this->getConnection());
            $group->setId($this->id);
            $group->addOwner($this->getConnection()->getUser());
        }

        // ensure the group has the new list of members and subgroups
        $group->setUsers((array) $this->getRawValue('members'));
        $group->setSubgroups((array) $this->getRawValue('subgroups'));

        // if this is an edit and we're an owner pass editAsOwner = true to allow admin access
        // if this is an add and we're not a super user pass addAsAdmin = true to improve our chances
        $group->save(
            !$isAdd && in_array($this->getConnection()->getUser(), $group->getOwners()),
            $isAdd  && !$this->getConnection()->isSuperUser()
        );

        // invalidate group cache because we know groups have changed.
        // @todo    this will happen again due to the group trigger (which is inefficient)
        //          we do it here anyway because the trigger can take a moment to propagate
        //          and we don't want the user to see stale data.
        try {
            $cache = $this->getConnection()->getService('cache');
            $cache->invalidateItem('groups');
        } catch (ServiceNotFoundException $e) {
            // no cache? nothing to invalidate.
        }

        parent::save();

        // clear cache and return
        static::clearCache($this->getConnection());
        return $this;
    }

    /**
     * Tests whether the given user is a member of this project.
     *
     * @param   string  $userId     ID of the user we are checking membership
     * @return  bool    whether or not the user is a member of this project
     * @throws \P4\Exception
     */
    public function isMember($userId)
    {
        if (!$userId) {
            return false;
        }

        $members = $this->getAllMembers(true);
        return isset($members[$userId]);
    }

    /**
     * Tests whether the given user is an owner of this project.
     *
     * @param   string  $userId     ID of the user we are checking ownership
     * @return  bool    whether or not the user is an owner of this project
     */
    public function isOwner($userId)
    {
        if (!$userId) {
            return false;
        }
        $owners = $this->getOwners();
        return in_array($userId, $owners);
    }

    /**
     * Tests whether the given user is following this project.
     *
     * @param   string  $userId     ID of the user we are checking
     * @return  bool    whether or not the user is a follower of this project
     * @throws \P4\Exception
     */
    public function isFollowing($userId)
    {
        if (!$userId) {
            return false;
        }
        $members  = $this->getAllMembers();
        $follower = $this->getFollowers($members);
        return in_array($userId, $follower);
    }
    
    /**
     * Tests whether the given user is a moderator of this project.
     *
     * @param   string  $userId     ID of the user we are checking
     * @param   array|null  $branches   optional - limit branches to collect moderators from
     * @return  bool    whether or not the user is a moderator of this project
     * @throws \P4\Exception
     */
    public function isModerator($userId, array $branches = null)
    {
        if (!$userId) {
            return false;
        }
        $moderators = $this->getModerators($branches);
        return in_array($userId, $moderators);
    }

    /**
     * Gets Membership levels of this project.
     *
     * @param   string  $userId     ID of the user we are checking the membership
     * @param   bool    $asArray    Whether to return results as an array or imploded string
     * @return  string  list of membership
     * @throws \P4\Exception
     */
    public function getMembershipLevels($userId, $asArray = false)
    {
        if (!$userId) {
            return false;
        }
        $membershipLevels = array();
        if ($this->isOwner($userId)) {
            array_push($membershipLevels, 'Owner');
        }
        if ($this->isMember($userId)) {
            array_push($membershipLevels, 'Member');
        }
        if ($this->isFollowing($userId)) {
            array_push($membershipLevels, 'Following');
        }
        if ($this->isModerator($userId)) {
            array_push($membershipLevels, 'Moderator');
        }

        return $asArray === true ? $membershipLevels : implode(", ", $membershipLevels);
    }

    /**
     * Indicate whether a user has any of the given membership level for a project.
     *
     * @param   User    $user       User object to query
     * @param   array   $roles      A list of roles to check for involvement
     * @return  bool    true/false  involved/not involved
     * @throws \P4\Exception
     */
    public function isInvolved(User $user, $roles = null)
    {
        $involved = false;
        if ($user) {
            $id           = $user->getId();
            $rolesToCheck = Project::$projectRoles;
            if ($roles) {
                $rolesToCheck = array_intersect($rolesToCheck, $roles);
            }
            // Check members/owners first so we can exit early if found
            $involved = in_array(Project::MEMBERSHIP_LEVEL_MEMBER, $rolesToCheck) && $this->isMember($id) ||
                        in_array(Project::MEMBERSHIP_LEVEL_OWNER,  $rolesToCheck) && $this->isOwner($id);
            if (!$involved) {
                $follows  =
                    in_array(Project::MEMBERSHIP_LEVEL_FOLLOWER, $rolesToCheck)
                        ? $user->getConfig()->getFollows('project')
                        : null;
                $involved =
                    $follows && in_array($this->getId(), $follows) ||
                    (in_array(Project::MEMBERSHIP_LEVEL_MODERATOR, $rolesToCheck) && $this->isModerator($id));
            }
        }
        return $involved;
    }

    /**
     * Gets Membership level of this project for sorting.
     *
     * @param   string  $userId     ID of the user we are checking the membership
     * @return  int     level for sorting
     * @throws \P4\Exception
     */
    public function getMembershipLevelForSort($userId)
    {
        if ($this->isOwner($userId)) {
            return 1;
        } elseif ($this->isMember($userId)) {
            return 2;
        } elseif ($this->isFollowing($userId)) {
            return 3;
        } elseif ($this->isModerator($userId)) {
            return 4;
        }
        return 0;
    }

    /**
     * Fetch project group to populate members and subgroups if needed.
     *
     * On 2012.1+ p4d instances members and subgroups are stored in a group
     * with the id swarm-project-<projectId>. On servers older than 2012.1
     * admin users are unable to manage groups so members are stored in the
     * project key and subgroups are not supported.
     */
    protected function loadGroup()
    {
        $setMembers   = !array_key_exists('members',   $this->values) && $this->needsGroup;
        $setSubgroups = !array_key_exists('subgroups', $this->values) && $this->needsGroup;

        if (!$setMembers && !$setSubgroups) {
            return;
        }

        $members   = array();
        $subgroups = array();
        try {
            $group     = Group::fetchById($this->id, $this->getConnection());
            $members   = $group->getUsers();
            $subgroups = $group->getSubgroups();
        } catch (\P4\Spec\Exception\NotFoundException $e) {
        } catch (\InvalidArgumentException $e) {
        }

        $this->setRawValue('members',   $setMembers   ? $members   : $this->getRawValue('members'));
        $this->setRawValue('subgroups', $setSubgroups ? $subgroups : $this->getRawValue('subgroups'));
        $this->needsGroup = false;
    }

    /**
     * Get projects from the cache. If the cache is empty, populate it.
     *
     * @param   Connection  $p4     specific connection to use
     * @return  Iterator            cached project records
     * @throws \Exception
     */
    protected static function getCachedProjects(Connection $p4)
    {
        // get models from cache if available
        $cache    = $p4->getService('cache');
        $projects = $cache->getItem('projects');

        // if cache is empty, build it
        if ($projects === null) {
            $projects = static::fetchAll(
                array(
                    static::FETCH_NO_CACHE        => true,
                    static::FETCH_INCLUDE_DELETED => true
                ),
                $p4
            );
            $cache->setItem('projects', $projects);
        }

        return $projects;
    }

    /**
     * Extends parent to flag models as requiring a group populate if
     * the server version is new enough to make it possible for admins.
     *
     * @param   Key             $key        the key to record'ize
     * @param   string|callable $className  ignored
     * @return  AbstractKey     the record based on the passed key's data
     */
    protected static function keyToModel($key, $className = null)
    {
        $model             = parent::keyToModel($key);
        $model->needsGroup = true;
        return $model;
    }

    /**
     * Check if a readme file exists and process the file to be displayed if found.
     * @param  array     $readmeLocations    Location(s) to check if there is a readme.md file
     * @param  int       $maxSize            Max size to be displayed
     * @param  mixed     $markdownHelper     Markdown helper provider.
     * @param  array     $readmeExtensions   File extensions to consider. Defaults to 'md' if empty
     * @return string    readme              The data needed to render the readme
     * @throws \P4\Exception
     * @throws \P4\File\Exception\Exception
     * @throws \P4\File\Exception\NotFoundException
     */
    public function getReadme($readmeLocations, $maxSize, $markdownHelper, $readmeExtensions)
    {
        $project  = $this;
        $p4Admin  = $this->getConnection();
        $branches = $project->getBranches('id', $readmeLocations);

        // Check each mainline branch from the config to get a readme file
        // if multiple matching branches - get the file from the first one created
        $readme = false;
        // In case array_column doesn't exist use get_field_data
        if (function_exists('array_column')) {
            $branchNames = array_column($branches, 'id');
        } else {
            $branchNames = ArrayValues::getFieldData($branches, 'id');
        }
        foreach ($readmeLocations as $line) {
            if (in_array($line, $branchNames)) {
                $branch = array_filter(
                    $branches,
                    function ($elem) use ($line) {
                        return $elem['id'] === $line;
                    }
                );
                $branch = reset($branch);
                foreach ($branch['paths'] as $depotPath) {
                    // Break early if we have exclusion mappings
                    if (substr($depotPath, 0, 1) === '-') {
                        break;
                    }
                    $readmeExtensions = empty($readmeExtensions) ? ['md'] : $readmeExtensions;
                    //  Remove ... from the end of a depot path
                    $filePath       = substr($depotPath, -3) === '...' ? substr($depotPath, 0, -3) : null;
                    $extensionRegex = implode('|', $readmeExtensions);
                    // filter is case insensitive and if file is not deleted/renamed.
                    $filter = Filter::create()->add(
                        'depotFile',
                        $filePath . 'readme.(' . $extensionRegex . ')$',
                        Filter::COMPARE_REGEX,
                        Filter::CONNECTIVE_AND,
                        true
                    )->add(
                        'headAction',
                        'delete',
                        Filter::COMPARE_NOT_CONTAINS,
                        Filter::CONNECTIVE_AND,
                        true
                    );

                    $query = Query::create()->setFilter($filter);
                    $query->setFilespecs($filePath ? $filePath . '*' : $depotPath);

                    $fileList = File::fetchAll($query);
                    // There may be multiple files present, use the order of the extensions to pick
                    // the preferred file
                    foreach ($readmeExtensions as $extension) {
                        foreach ($fileList as $file) {
                            $ext = pathinfo($file->getFileSpec(), PATHINFO_EXTENSION);
                            if ($ext === $extension) {
                                $readme = File::fetch($file->getFileSpec(), $p4Admin, true);
                                break(4);
                            }
                        }
                    }
                }
            }
        }

        if ($readme) {
            $purifiedMarkdown = $markdownHelper->get(ViewHelperFactory::MARKUP);
            $maxReadMeSize    = $maxSize;

            // If file is over $maxReadMeSize it will display upto that size and cut any ending.
            $contents = $readme->getDepotContents(
                array(
                    $readme::UTF8_CONVERT  => true,
                    $readme::UTF8_SANITIZE => true,
                    $readme::MAX_SIZE      => $maxReadMeSize
                )
            );

            $readme = $purifiedMarkdown($contents);
        } else {
            $readme = '';
        }
        return $readme;
    }

    /**
     * Gets the project part of 'swarm-project-xxx'
     * @param $projectName the project name
     * @return mixed the stripped project if it is a project or the value unchanged if not
     */
    public static function getProjectName($projectName)
    {
        return preg_replace('/^' . Project::KEY_PREFIX . '/', '', $projectName);
    }

    /**
     * Tests to see if the name starts with 'swarm-project-'
     * @param $projectName the project name
     * @return bool true if the name is a Swarm project name
     */
    public static function isProjectName($projectName)
    {
        return Project::getProjectName($projectName) !== $projectName;
    }

    /**
     * Gets the workflow id defined at the project level or on a branch if the branch id is provided.
     * A workflow is optional so this may return null when no workflow is set
     * @param string            $branchId       Optional value. If provided and the branch is found
     *                                          the workflow for the branch will be returned.
     *                                          If a workflow is not set on the branch it will fall back to the workflow
     *                                          at the project level (this may also be not set)
     * @return  string|null     the workflow id or null if the workflow was not found.
     */
    public function getWorkflow($branchId = null)
    {
        $workflowId = $this->getRawValue(Project::FIELD_WORKFLOW);
        if ($branchId) {
            $branch = $this->branchExists($branchId);
            if ($branch && isset($branch[Project::FIELD_WORKFLOW])) {
                $workflowId = $branch[Project::FIELD_WORKFLOW];
            }
        }
        return $workflowId;
    }

    /**
     * Links a project to a workflow by specifying the workflow id.
     * @param string    $workflowId     the id for the workflow to link to the project
     * @return mixed the project
     */
    public function setWorkflow($workflowId)
    {
        return $this->setRawValue(Project::FIELD_WORKFLOW, $workflowId);
    }

    /**
     * Sets whether default project level reviewers are retained.
     * @param bool  $retained   true to retain
     * @return AbstractKey
     */
    public function setRetainDefaultReviewers($retained)
    {
        return $this->setRawValue(Project::FIELD_RETAIN_DEFAULT_REVIEWERS, (bool)$retained);
    }

    /**
     * Gets whether default reviewers are retained
     * @param null|string $branchId     if provided the retention value from the branch is returned if the branch is
     *                                  found. Branch values are not inherited from the project
     * @return bool
     */
    public function areDefaultReviewersRetained($branchId = null)
    {
        $retained = false;
        if ($branchId) {
            $branch = $this->branchExists($branchId);
            if ($branch && isset($branch[Project::FIELD_RETAIN_DEFAULT_REVIEWERS])) {
                $retained = (bool)$branch[Project::FIELD_RETAIN_DEFAULT_REVIEWERS];
            }
        } else {
            $retained = (bool)$this->getRawValue(Project::FIELD_RETAIN_DEFAULT_REVIEWERS);
        }
        return $retained;
    }

    /**
     * Get the Minimum UpVotes required.
     *
     * @param null|string $branchId     if provided the minimum up vote value from the branch is returned if the
     *                                  branch is found. Branch values are not inherited from the project
     *
     * @return array|mixed|null
     */
    public function getMinimumUpVotes($branchId = null)
    {
        $minimumUpVotes = $this->getRawValue(Project::FIELD_MINIMUM_UP_VOTES);

        if ($branchId) {
            $branch = $this->branchExists($branchId);
            if ($branch && isset($branch[Project::FIELD_MINIMUM_UP_VOTES])) {
                $minimumUpVotes = $branch[Project::FIELD_MINIMUM_UP_VOTES];
            }
        }
        return $minimumUpVotes;
    }

    /**
     * Set the Minimum UpVotes required.
     * @param mixed $total This is the value that is going to be saved for the minimum up vote.
     *
     * @return AbstractKey
     */
    public function setMinimumUpVotes($total)
    {
        return $this->setRawValue(Project::FIELD_MINIMUM_UP_VOTES, $total);
    }
}
